"use strict";

const _ = require("lodash");



module.exports = {

  index (value, index) {
    if (arguments.length != 2) throw "index is required";
    if (_.isArray(value)) {
      this._.stack.push({key, value});
      return value[key];
    } else {
      throw "index expects an array";
    }
  },

  key (value, key) {
    if (arguments.length != 2) throw "key is required";
    if (_.isObject(value) || _.isArray(value)) {
      this._.stack.push({key, value});
      return value[key];
    } else {
      throw "key expects an array or object";
    }
  },

  parent () {
    if (!this._.stack.length) return;
    return this._.stack.pop().value;
  },

  sibling (value, key) {
    if (arguments.length != 2) throw "key is required";
    if (!this._.stack.length) return;
    let s = this._.stack.pop();
    s.key = key;
    this._.stack.push(s);
    return s.value[s.key];
  },

  rename (value, newKey) {
    if (!this._.stack.length) return;
    let s = this._.stack.pop();
    if (_.isObject(s.value)) {
      delete s.value[s.key];
      s.value[newKey] = value;
      s.key = newKey;
      this._.stack.push(s);
      return value;
    } else {
      throw "rename expects the parent to be an object";
    }
  },

  duplicate (value, newKey) {
    if (!this._.stack.length) return value;
    let parent = this._.stack[this._.stack.length - 1].value;
    if (_.isObject(parent)) {
      delete s.value[s.key];
      s.value[newKey] = value;
      s.key = newKey;
      this._.stack.push(s);
      return value;
    } else {
      throw "duplicate expects the parent to be an object";
    }
  },

  freeze (value) {
    if (_.isObject(value)) {
      return Object.freeze(value);
    } else {
      throw "freeze expects an object";
    }
  },

  empty () {
    if (!this._.stack.length) return;
    let s = this._.stack[this._.stack.length];
    if (_.isArray(s.value) || _.isObject(s.value)) {
      s.value[s.key] = undefined;
    }
  },

  delete () {
    if (!this._.stack.length) return;
    let s = this._.stack.pop();
    if (_.isArray(s.value)) {
      s.value.splice(s.key, 1);
      return s.value;
    } else if (_.isObject(s.value)) {
      delete s.value[s.key];
      return s.value;
    } else {
      throw "Cannot delete";
    }
  }

};